#######################################################################
#
# Directory.pm - Perl module for searching the Stanford Directory
#
# $Id: Directory.pm,v 1.37 2005/08/09 21:29:04 quanah Exp $
#
#----------------------------------------------------------------------
#  Copyright  2000-2004 Board of Trustees, Leland Stanford Jr. University
#
#  This module is free software; you may redistribute it and/or modify it
#  under the same terms as Perl itself. 
#######################################################################

package Stanford::Directory;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK $AUTOLOAD);

use Carp;
use Net::LDAP;
use MIME::Base64;
use Authen::SASL qw(Cyrus);
use Socket;
use Stanford::Directory::Entry;
use Authen::Krb5;

require Exporter;

@ISA = qw(Exporter);

# Global SASL context.  This is workaround to overcome initialization
# problems with Net::LDAP, Authen::SASL, and Authen::Krb5 in the same
# context.
our $SASL;

#------------------------------------------------------------------------
# new (constructor) - set up directory connection parameters

sub new {
    my $that  = shift;
    my $class = ref($that) || $that;

    my %fields =
        (
         ldap_conn      => undef,
         ldap_server    => "ldap.stanford.edu",
         mechanism      => undef,
         version        => 3,
         scope          => "sub", 
         service        => undef,
         instance       => undef,
         keytab         => undef,
         basedn         => "dc=stanford, dc=edu",
         ticket_file    => "/tmp/directory.$<.tkt",
         error          => "",
         error_type     => "",
         keytab_init    => 0,
         debug          => 0
        );

    my $self  = {
        _permitted => \%fields,
        %fields,
    };
    bless $self, $class;
    return $self;
}

#------------------------------------------------------------------------
# AUTOLOAD - used to get and set parameters individually

sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self);
    
    my $name = $AUTOLOAD;
    $name =~ s/.*://;               # strip fully-qualified portion
    return if $name =~ /DESTROY/;   # don't do DESTROY
    
    unless (exists $self->{_permitted}->{$name}) {
        carp "$type doesn't have a `$name'";
        return undef;
    }
    
    if (@_) { $self->{$name} = shift }
    return $self->{$name};
}

#------------------------------------------------------------------------
# set - set parameters en mass

sub set {
    my $self = shift;
    my $type = ref($self);
    my $name;
    while ($name = shift) {
        unless (exists $self->{_permitted}->{$name}) {
            carp "$type doesn't have a `$name'";
            shift;
            next;
        }
        $self->{$name} = shift;
    }
}

#------------------------------------------------------------------------
# as_string - just a toy to list a Directory for debugging and such

sub as_string {
    my $self = shift;
    my $type = ref($self) or croak "$self is not an object";
    my $string;

    foreach (sort keys %{$self->{_permitted}}) {
        $string .= sprintf("%16s : %s\n", $_, $self->{$_});
    }
    return $string;
}

#------------------------------------------------------------------------
# Initialize SASL

sub sasl_init {
    my $self = shift;
    if (!$SASL) {
        print "Creating SASL object with mechanism ".$self->mechanism()."\n" 
            if $self->debug;
        $SASL = Authen::SASL->new($self->mechanism());
    }
    return;
}

#------------------------------------------------------------------------
# ldap_query - Using the given filter, contact the LDAP server and see
#              what we get back.  Parse it into a list of Entries, one
#              for each record returned.

sub ldap_query {
    my ($self, $filter, @attrs) = @_;

    # Clear error, error_type
    $self->error("");  
    $self->error_type("");

    # Check to see if there is more than one filter
    my $singlef=0;
    if ( !ref $filter) {
        $filter = [ $filter ];
        $singlef = 1;
        print "Processing single filter\n" if $self->debug;
    }

    # Connect to the directory
    $self->open_directory();
    my $ld = $self->ldap_conn;
    die $self->error_type.': '.$self->error."\n" if $self->error;

    # Perform the query
    my %entries;
    foreach (@$filter) {
        my @entries;
        print "Performing search with filter $_\n" if $self->debug;
        my $status = $ld->search(base=>$self->basedn, 
                                 scope=>$self->scope, 
                                 filter=>$_, 
                                 attrs=>\@attrs);

        # Check for errors
        if ( $status->is_error ) {
            my $err;
            $err .= $status->error_name().' - '.$status->error()."\n";
            $err .= " (try narrowing your search)" if ($err =~ /exceeded$/);
            $self->error_type("Search");
            $self->error($err);
            return () unless ($err =~ /exceeded$/);
        }

        # No errors, cool. Grab the entries make them Entries.
        my @entrys;
        my $ent;
        @entrys=$status->all_entries;
        foreach $ent (@entrys) {
            my $entry = Stanford::Directory::Entry->new();
            $entry->dn($ent->dn);
            foreach my $attr ($ent->attributes()) {
                my $key = lc($attr);
                $entry->$key($ent->get_value($attr));
            }
            push (@entries, $entry);
        }
        if ($singlef == 1) {
            return @entries;
        } else {
            $entries{$_}=\@entries;
        }
    }

    return %entries;
}

#------------------------------------------------------------------------
# This function is here for backwards compatibility and there is no
# reason to call it any more.  This query function was originally
# added as a performance improvement and is not really needed anymore
# since the LDAP connections are kept until directory_close is
# explicitly called.

sub ldap_persistent_query {
    my ($self, $filter, @attrs) = @_;
    my @entries = $self->ldap_query($filter, @attrs);
    return @entries;
}

#------------------------------------------------------------------------
# manage_ksrvtgt - make sure we have a valid ticket (this isn't really
#                  a method, because it might not really belong here).

sub _manage_ksrvtgt {
    my ($user, $instance, $keytab, $file) = @_;

    # Check for keytab
    return "Cannot read keytab file, $keytab" unless (-r $keytab);

    my $status;
    {
         Authen::Krb5::init_context();
         Authen::Krb5::init_ets();
         my $client = Authen::Krb5::parse_name("$user/$instance");
         my $server = Authen::Krb5::parse_name("krbtgt/stanford.edu");
         my $cc = Authen::Krb5::cc_resolve("FILE:$file");
         $cc->initialize($client);
         my $kt = Authen::Krb5::kt_resolve($keytab);
         Authen::Krb5::get_in_tkt_with_keytab($client,$server,$kt,$cc)
           or die Authen::Krb5::error() . " while getting Kerberos ticket\n"; 
    }
    
    # Set environment variable.
    $ENV{KRB5CCNAME} = "FILE:$file";

    return 1 unless $status; # success
    return "Cannot guarantee ticket for \"$user.$instance\" (status $status)";
}

#------------------------------------------------------------------------
# Open a connection to the directory

sub open_directory {
    my ($self) = @_;
    
    # We only need to connect once
    return if $self->ldap_conn;

    # Connect to the server
    my $serverName = gethostbyaddr(inet_aton($self->ldap_server()), AF_INET);
    my $ld = Net::LDAP->new($serverName,version=>$self->version());
    if ($ld == -1) {
        $self->error_type("System");
        $self->error("Could not connect to directory server (" 
                     . $self->ldap_server .")");
        return ();
    }

    # Manage the tgt if we've got a keytab --
    #   This should be abstracted into something like $self->authenticate
    if ($self->keytab) {
        my $err = _manage_ksrvtgt ($self->service,
                                   $self->instance,
                                   $self->keytab,
                                   $self->ticket_file,
            );
        if ($err ne "1") {
            $self->error_type("System");  $self->error($err);
            return ();
        }
    }

    # Bind to the directory
    my $status;
    if ($self->mechanism) {
        $self->sasl_init();
        print "SASL bind to the directory\n" if $self->debug;
        $status = $ld->bind("", sasl=>$SASL);
    } else {
        $status = $ld->bind("");
    }

    if ($status->code && $status->error) {
        $self->error_type("System");
        $self->error("Bind failed - " . $status->error 
                     . " (STATUS: " . $status->code . ")");
        return ();
    }
     $self->ldap_conn($ld);
}

#------------------------------------------------------------------------
# Close the open directory connection

sub close_directory {
    my ($self) = @_;
    my $ld = $self->ldap_conn;
    $ld->unbind if $ld;
}

1;
__END__

=head1 NAME

Stanford::Directory - Perl module for searching the Stanford Directory

=head1 SYNOPSIS

use Stanford::Directory;

$dir = new Stanford::Directory;

$dir->set(attribute =E<gt> 'value', ...);

@entries = $dir->ldap_query($filter [, @attributes ]);
or
@entries = $dir->ldap_query(\@filter [, @attributes ]);

=head1 PERSISTENT SYNOPSIS

use Stanford::Directory;

$dir = new Stanford::Directory;

$dir->set(attribute =E<gt> 'value', ...);

$dir->open_directory;


@entries = $dir->ldap_persistent_query($filter [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter2 [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter3 [, @attributes ]);
@entries = $dir->ldap_persistent_query($filter4 [, @attributes ]);
...

$dir->close_directory;

=head1 DESCRIPTION

B<Stanford::Directory> is a light-weight object for accessing the
Stanford Directory.  Use it to maintain the information necessary
to connect to the directory and to search the directory using LDAP
filters.

=head1 METHODS

B<new>

Allocates and returns a new B<Stanford::Directory> object.

B<set(attribute =E<gt> 'value', ...)>

Use I<set> to set the various B<Stanford::Directory> attributes (see
the B<ATTRIBUTES> section below).  Each attribute can also be set or
retrieved by using its name as a method.

B<ldap_query($ldap_filter [, @attributes ])>
B<ldap_query(\@ldap_filter [, @attributes ])>

The I<ldap_query> method makes a directory search using the supplied
I<ldap_filter>.  It returns an array of B<Stanford::Directory::Entry>
objects.  The I<attributes> array contains the names of directory
entry attributes to include in each B<Stanford::Directory::Entry>.  If
it is empty or omitted all of the directory entry attributes are
included.  Alternately, you can provide a reference to an array of
filters to supply to the I<ldap_query> method.  The benefit of using
an array of filters is that multiple searches will be done inside a
single bind to the directory server, which is much speedier than doing
multiple single-filter queries.  It will return a hash of references
to B<Stanford::Directory::Entry> objects that are keyed by search
filter.

B<open_directory>

Opens up a persistent connection to the directory.

B<ldap_persistent_query($ldap_filter [, @attributes])>

The I<ldap_persistent_query> method behaves the same way as the
I<ldap_query> method, except that it does not close the connection
when it finishes the search.  Subsequent searches are done using the
same connection.  If the connection gets closed between searches,
I<ldap_persistent_query> will open a new one, and then resume
searching.

This method is obsolete.  The ldap_query method no longer closes the
connection when it completes.

B<close_directory>
Close a persistent connection to the directory.

=head1 ATTRIBUTES

The B<Stanford::Directory> object has the following attributes (with
default values, if any):

  ldap_server     LDAP server name (ldap.stanford.edu)
  mechanism       undef (for anonymous binds) or GSSAPI
  service         Kerberos service principal (service, webauth, etc)
  instance        Kerberos instance
  keytab          keytab location for K5
  basedn          Search base DNS (dc=stanford, dc=edu)
  scope           Search scope (one,base,sub) default: sub
  ticket_file     Kerberos ticket file location (/tmp/directory.UID.tkt)
  error           Error text
  error_type      Error type - "System" or "Search"
  debug           Defaults to 0.  If set then debugging messages are printed.

B<Stanford::Directory> now supports multiple authentication types.
SASL mechanisms and anonymous binding are available.  If you supply a
I<keytab> and a I<service>, B<Stanford::Directory> will manage the
kerberos tgt for you.  You can also specify where you want the
kerberos ticket stored by using I<ticket_file>.

I<error> and I<error_type> are set whenever an error occurs.

=head1 EXAMPLE for Kerberos

  use Stanford::Directory;

  $dir = new Stanford::Directory;
  $dir->set(ldap_server    => "ldap.stanford.edu",
            mechanism      => "GSSAPI",
            basedn         => "cn=people, dc=stanford, dc=edu");

  $s = shift or die "usage: $0 <string>\n";
  $filter = "(sn=$s)";

  @entries = $dir->ldap_query($filter,"displayname");
  if (!@entries) {
      if ($dir->error) {
          print "\n" . $dir->error_type .
                " Error:\n\n   " . $dir->error . "\n\n";
      } else {
          print "No match.\n";
      }
      exit;
  }

  print "People with last names like '$s':\n\n";
  foreach (@entries) { print "  ", $_->displayname, "\n"; }

=head1 EXAMPLE for Kerberos and an array of filters

  use Stanford::Directory;

  $dir = new Stanford::Directory;
  $dir->set(ldap_server    => "ldap.stanford.edu",
            mechanism      => "GSSAPI",
            basedn         => "cn=people, dc=stanford, dc=edu");

  $s = shift or die "usage: $0 <string>\n";
  @filter = ("sn=$s", "cn=$s");

  %entries = $dir->ldap_query(\@filter,"displayname");
  if (!%entries) {
      if ($dir->error) {
          print "\n" . $dir->error_type .
                " Error:\n\n   " . $dir->error . "\n\n";
      } else {
          print "No match.\n";
      }
      exit;
  }

  print "People with last names like '$s':\n\n";
  foreach $key (keys %entries) { 
      foreach (@{$entries{$key}}) {
          print "  ", $_->displayname, "\n";
      }
   }

=head1 EXAMPLE for Kerberos and an persistent connection

  use Stanford::Directory;
  
  $dir = new Stanford::Directory;
  $dir->set(ldap_server    => "ldap.stanford.edu",
            mechanism      => "GSSAPI",
            basedn         => "cn=people, dc=stanford, dc=edu");

  @users=`cat /afs/ir/service/etc/passwd`;
  
  $dir->open_directory;

  foreach $user (@users) {
    ($uid, $rest)=split /:/, $user, 2;
    @entries = $dir->ldap_persistent_query("uid=$uid","displayname");
    if (!@entries) {
        if ($dir->error) {
            print "\n" . $dir->error_type .
                  " Error:\n\n   " . $dir->error . "\n\n";
        } else {
            print "No match.\n";
        }
        exit;
    }

    $entry = shift (@entries);
    print "displayname for $uid is " . ($entry->displayname)[0] . "\n";
  }

  $dir->close_directory;

=head1 SEE ALSO

Net::LDAP, Stanford::Directory::Entry, and example scripts provided
in the I<examples> directory.

=head1 AUTHORS

Rob Riepel, riepel@stanford.edu

Quanah Gibson-Mount, quanah@stanford.edu

=head1 COPYRIGHT AND LICENSE

    Copyright 2000-2004 Board of Trustee, Leland Stanford Jr. University

    This module is free software; you may redistribute it and/or modify it
    under the same terms as Perl itself.

=cut
